import jcmwave
import numpy as np


class ForwardProblem:
    def __init__(self, project_file, wavelengths, material):
        self._wavelengths = wavelengths
        self._material = material

        self._project_file_str = str(project_file.absolute())

    def solve(self, keys):
        results, logs = self._run_project(keys)
        matrices, matrices_derivatives = self._mueller_matrix(results)

        return matrices, matrices_derivatives

    def _run_project(self, keys):

        job_ids = list()

        for wavelength in self._wavelengths:

            # Get a local copy of the keys for modifying and dispatching
            inner_keys = keys.copy()
            inner_keys["vacuum_wavelength"] = wavelength * 1e-9
            inner_keys["n3"] = self._material.nk(wavelength)
            job_id = jcmwave.solve(self._project_file_str, inner_keys, temporary=True)
            job_ids.append(job_id)

        # here we wait until all jobs are finished
        results, logs = jcmwave.daemon.wait(job_ids=job_ids)

        return results, logs

    def _mueller_matrix(self, results):

        mueller_matrices = list()
        mueller_matrices_derivatives = dict()

        # Iterate over all results
        for result in results:

            # First get the Mueller matrix
            M = np.zeros((4, 4))
            for i in range(4):
                for j in range(4):
                    M[i, j] = result[2]["Mueller_ps{:1d}{:1d}".format(i + 1, j + 1)][0][
                        0
                    ]
            mueller_matrices.append(M)

        # Cast to numpy and flatten
        mueller_matrices = np.asarray(mueller_matrices)
        mueller_matrices = mueller_matrices#.flatten()

        # Then obtain the derivatives of the Mueller matrix wrt the model parameters; if
        # keys are not found no error is thrown
        for i, parameter in enumerate(["h", "width", "swa", "radius"]):
            derivative_key = "d_{}".format(parameter)
            parameter_derivative = list()
            for result in results:
                if derivative_key in result[2]:

                    dM = np.zeros((4, 4))
                    for i in range(4):
                        for j in range(4):
                            dM[i, j] = result[2][derivative_key][
                                "Mueller_ps{:1d}{:1d}".format(i + 1, j + 1)
                            ][0][0]
                    parameter_derivative.append(dM)

            parameter_derivative = np.asarray(parameter_derivative)
            parameter_derivative = parameter_derivative#.flatten()

            mueller_matrices_derivatives[parameter] = parameter_derivative

        # Returns a list with the Mueller matrices M = [M1, M2, ...] (each 4x4) and a list
        # with derivatives dM = [dM1dh, dM2dh, ... , dM1dwidth, dM2dwidth, ...] (each 4x4)
        return mueller_matrices, mueller_matrices_derivatives
